'''
Subpackage myimg.utils
----------------------

The subpackage with code for specific/more complex utilities in myimg:

* myimg.utils.label = draw a label in the upper left corner of an image
* myimg.utils.caption = add a one-line caption at the bottom  of an image
* myimg.utils.scalebar = draw a scalebar in a micrograph (or diffractogram)
'''
